package jig.engine.lwjgl;

import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;

import jig.engine.ImageResource;
import jig.engine.RenderingContext;

import org.lwjgl.opengl.GL11;

/**
 * This object is responsible for keeping track of a given OpenGL texture 
 * and for calculating the texturing mapping coordinates of the full image.
 * 
 * Texture widths and heights need to be a power of two (except in the most
 * recent OpenGL versions), so the actual texture may be larger than the source
 * image. This class helps to manage these two sets of dimensions, and is also
 * used to render the texture onto a quad.
 *
 * Portions of this code were borrowed from the LWJGL project's example code.
 * 
 * @author Andrew Nierman
 */
class Texture implements ImageResource {
	
    private int textureID;
    // The height of the original image
    private int height;
    // The width of the original image
    private int width;
    
    // The width of the texture, must be a power of two
    private int texWidth;
    // The height of the texture, must be a power of two
    private int texHeight;
    
    // The ratio of the width of the image to the texture width
    @SuppressWarnings("unused")
	private double widthRatio;
    // The ratio of the height of the image to the texture height
    @SuppressWarnings("unused")
	private double heightRatio;
    
    /**
     * Create a new texture, based on an image with a given width and height.
     * The OpenGL texture will have dimensions that are at least as large as
     * these values, since OpenGL textures need to be powers of two.
     *
     * @param textureID The OpenGL texture ID generated by glGenTextures
     * @param imageWidth The width of the original image
     * @param imageHeight The height of the original image
     */
    public Texture(final int textureID,
    		final int imageWidth, final int imageHeight) {
        this.textureID = textureID;
        this.width = imageWidth;
        this.height = imageHeight;

        this.texWidth = dimensionAsPowerOfTwo(imageWidth);
        this.texHeight = dimensionAsPowerOfTwo(imageHeight);

    	heightRatio = ((double) height) / texHeight;
    	widthRatio = ((double) width) / texWidth;
    }
    
    /**
     * Bind this texture so that it has the current context.
     *
     */
    private void bind() {
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, textureID);
    }
    
    /**
     * Get the height of the original image.
     *
     * @return The height of the original image
     */
    public int getHeight() {
        return height;
    }
    
    /** 
     * Get the width of the original image.
     *
     * @return The width of the original image
     */
    public int getWidth() {
        return width;
    }
    
    /**
     * Get the width of the texture.
     * 
     * @return The width of the texture as a power of two.
     */
    public int getTextureWidth() {
    	return texWidth;
    }
    
    /**
     * Get the height of the texture.
     * 
     * @return The height of the texture as a power of two.
     */
    public int getTextureHeight() {
    	return texHeight;
    }
    
    /**
	 * Returns the smallest number greater or equal to actualDimension,
	 * which is also a power of two, since texture dimensions in OpenGL are
	 * required to be a power of two (except in the more recent versions).
	 * Note that the "red book" also indicates that the texture must be at
	 * least 64x64 in its documentation of the glTexImage2D function,
	 * but I have not found this to be the case.
     * 
     * @param actualDimension The size of the actual image dimension
     * @return The size of the corresponding texture dimension
     */
    private int dimensionAsPowerOfTwo(final int actualDimension) {
    	//int textureDimension = 64;
        int textureDimension = 2;
        
        while (textureDimension < actualDimension) {
            textureDimension *= 2;
        }
        
        return textureDimension;
    }
    
    /**
     * Maps the bound texture onto a quad, resulting in the rendering of the
     * image.
     * 
     * @param rc the current rendering context
     * @param at the transformation to apply to the sprite/image
     * 
     */
    public void render(final RenderingContext rc, final AffineTransform at) {
		GL11.glPushMatrix();
		
		// bind to the appropriate texture
		bind();
    
    	rc.transform(at);
		
		// draw a quad textured with the image
    	GL11.glBegin(GL11.GL_QUADS);

    	GL11.glTexCoord2d(0, 0);
		GL11.glVertex2d(0, 0);
		GL11.glTexCoord2d(1.0, 0);
		GL11.glVertex2d(getWidth(), 0);
		GL11.glTexCoord2d(1.0, 1.0);
		GL11.glVertex2d(getWidth(), getHeight());
		GL11.glTexCoord2d(0, 1.0);
		GL11.glVertex2d(0, getHeight());
		
		GL11.glEnd();
		
		// restore the model view matrix
		GL11.glPopMatrix();		
    }
    /**
	 * TODO: Incomplete!
	 */
	public void draw(Graphics2D g, AffineTransform at) {
		
	}
}
